﻿Option Strict On
Option Infer On

Imports System.IO
Imports System.Threading
Imports DirectShow

Friend Class SoundEffect

    Private player As MediaPlayer
    Private unloaded As Boolean
    Private startPlay As Boolean
    Private stopPlay As Boolean
    Private exitOnStop As Boolean

    Private Shared m_GlobalVolume As Integer
    Private Shared t As Thread
    Private Shared sounds_ As List(Of SoundEffect)
    Private Shared exitThread As Boolean
    Private Shared locker As Object

    Friend Event playing(ByVal sender As Object, ByVal e As EventArgs)
    Friend Event stopped(ByVal sender As Object, ByVal e As EventArgs)

    ''' <summary>
    ''' Die Lautstärke aller Soundeffekte.
    ''' </summary>
    ''' <value></value>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Friend Shared Property globalVolume As Integer
        Get
            Return m_GlobalVolume
        End Get
        Set(value As Integer)
            m_GlobalVolume = value
        End Set
    End Property

    Private Shared ReadOnly Property sounds As List(Of SoundEffect)
        Get
            SyncLock locker
                Dim l As New List(Of SoundEffect)(sounds_)
                Return l
            End SyncLock
        End Get
    End Property

    Private Shared Sub addSound(se As SoundEffect)
        SyncLock locker
            sounds_.Add(se)
        End SyncLock
    End Sub

    Private Shared Sub removeSound(se As SoundEffect)
        SyncLock locker
            sounds_.Remove(se)
        End SyncLock
    End Sub

    ''' <summary>
    ''' Initialisiert das Soundmanagement.
    ''' </summary>
    ''' <remarks>Muss aufgerufen werden, um Soundeffekte wiedergeben zu können.</remarks>
    Friend Shared Sub init()

        sounds_ = New List(Of SoundEffect) 'Auflistung aller Soundeffekte

        locker = New Object

        m_GlobalVolume = 1

        'Hintergrundthread vorbereiten
        exitThread = False
        t = New Thread(New ThreadStart(AddressOf backThread))
        t.Name = "soundeffect thread"
        t.IsBackground = True
        t.Start() 'starten

    End Sub

    Sub New(name As String)
        Me.New(name, False) 'erweiterten Konsruktor aufrufen
    End Sub

    Sub New(name As String, exitOnStop As Boolean)

        Me.exitOnStop = exitOnStop

        'Mediaplayer initialisieren
        unloaded = False
        player = New MediaPlayer 'Player erstellen
        player.Volume = 1
        player.MediaFile = Path.Combine(Game.getEnvironmentPath("sounds"), name) 'Datei laden

        'Ereignisse abfangen
        AddHandler player.Started, AddressOf onPlaying
        AddHandler player.Stopped, AddressOf onStopped

        'Instanz zur Liste hinzufügen
        addSound(Me)

    End Sub

    ''' <summary>
    ''' Führt den Messageloop für alle Soundeffekte aus.
    ''' </summary>
    ''' <remarks></remarks>
    Private Shared Sub backThread()

        'Messageloop
        Do Until exitThread 'bis Beendigung ausführen
            Thread.Sleep(10) 'Prozessorzeit sparen -> Latenz von 10ms

            'Soundeffektmanagement
            For Each s In sounds 'alle Soundeffekte enumerieren

                If s.player.Volume <> globalVolume Then s.player.Volume = globalVolume 'Lautstärke setzen

                'Abspielen
                If s.startPlay Then
                    s.player.Play()
                    s.startPlay = False
                End If

                'Stoppen
                If s.stopPlay Then
                    s.player.Stop()
                    s.stopPlay = False
                    If s.exitOnStop Then s.unload()
                End If

            Next
        Loop
    End Sub

    ''' <summary>
    ''' Startet die Audiowiedergabe.
    ''' </summary>
    ''' <remarks></remarks>
    Friend Sub play()

        If unloaded Then Throw New ObjectDisposedException("Dieser Soundeffekt wurde bereits entladen.") 'Soundeffekt nicht mehr aktiv
        startPlay = True 'Feld für Thread setzen

    End Sub

    ''' <summary>
    ''' Stoppt die Audiowiedergabe.
    ''' </summary>
    ''' <remarks></remarks>
    Friend Sub [stop]()

        If unloaded Then Throw New ObjectDisposedException("Dieser Soundeffekt wurde bereits entladen.") 'Soundeffekt nicht mehr aktiv
        stopPlay = True 'Feld für Thread setzen

    End Sub

    Private Sub onPlaying(sender As Object, e As EventArgs)
        RaiseEvent playing(Me, New EventArgs) 'Ereignis auslösen
    End Sub

    Private Sub onStopped(sender As Object, e As EventArgs)
        RaiseEvent stopped(Me, New EventArgs) 'Ereignis auslösen
    End Sub

    ''' <summary>
    ''' Entläd den Soundeffekt.
    ''' </summary>
    ''' <remarks>Der Soundeffekt ist danach nicht mehr aktiv.</remarks>
    Friend Sub unload()

        'Entladen
        If player.State <> MediaPlayerState.Stopped Then player.Stop() 'Player stoppen
        Do Until player.State = MediaPlayerState.Stopped : Thread.Sleep(1) : Loop
        unloaded = True 'Indikator setzen
        removeSound(Me) 'Instanz aus Auflistung entfernen

    End Sub

    ''' <summary>
    ''' Entläd alle aktiven Soundeffekte und beendet den Hintergrundthread.
    ''' </summary>
    ''' <remarks>Danch können keine Soundeffekte mehr abgespielt werden.</remarks>
    Friend Shared Sub unloadAll()

        'Hintergrundthread beenden
        exitThread = True
        Do Until t.ThreadState = ThreadState.Stopped : Thread.Sleep(1) : Loop 'auf Beendingung warten

        'Alle Soundeffekte entladen
        For Each sound In sounds 'Soundeffekte enumerieren
            sound.player.Stop() 'Player stoppen
            sound.unload() 'Soundeffekt entladen
        Next

    End Sub
End Class
